/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.schedule;

import com.google.common.collect.Lists;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.AsyncWorkQueueProviderService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.config.CentralConfigKey;
import org.artifactory.config.InternalCentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.mbean.MBeanRegistrationService;
import org.artifactory.schedule.mbean.ManagedExecutor;
import org.artifactory.spring.ContextReadinessListener;
import org.artifactory.spring.Reloadable;
import org.artifactory.state.model.ArtifactoryStateManager;
import org.artifactory.storage.db.DbService;
import org.artifactory.version.CompoundVersionDetails;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.jfrog.common.config.diff.DataDiff;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.ClassUtils;

import javax.annotation.Nonnull;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * @author yoavl
 */
@Service
@Reloadable(beanClass = TaskService.class,
        initAfter = {DbService.class, InternalCentralConfigService.class, ArtifactoryStateManager.class},
        listenOn = CentralConfigKey.none)
public class TaskServiceImpl implements TaskService, ContextReadinessListener, TaskServiceCallBack {
    private static final Logger log = LoggerFactory.getLogger(TaskServiceImpl.class);

    @Autowired
    private CachedThreadPoolTaskExecutor executor;
    @Autowired
    private AddonsManager addonsManager;

    private ConcurrentMap<String, InternalTask> activeTasksByToken = new ConcurrentHashMap<>();
    private ConcurrentMap<String, InternalTask> inactiveTasksByToken = new ConcurrentHashMap<>();

    @Override
    public void init() {
        java.util.concurrent.Executor concurrentExecutor = executor.getConcurrentExecutor();
        if (concurrentExecutor instanceof ArtifactoryConcurrentExecutor) {
            ContextHelper.get().beanForType(MBeanRegistrationService.class).
                    register(new ManagedExecutor((ArtifactoryConcurrentExecutor) concurrentExecutor),
                            "Executor Pools", "Task Service");
        }
    }

    @Override
    public void destroy() {
        ContextHelper.get().beanForType(AsyncWorkQueueProviderService.class).closeAllQueues();
        cancelAllTasks(true);
        //Shut down the executor service to terminate any async operations not managed by the task service
        executor.destroy();
    }

    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {
    }

    @Override
    public void reload(CentralConfigDescriptor oldDescriptor, List<DataDiff<?>> configDiff) {
    }

    @Override
    public void onContextReady() {
        activateTasks();
    }

    @Override
    public String startTask(Task task, boolean waitForRunning) {
        InternalTask internalTask = (InternalTask) task;
        String token = internalTask.getToken();
        ArtifactoryContext context = ContextHelper.get();
        if (context.isOffline()) {
            log.error("Artifactory is in offline state, task {} is not allowed to run during offline state!", token);
            return token;
        }
        // Verify can be started first
        canBeStarted(internalTask);
        boolean contextReady = context.isReady();
        ConcurrentMap<String, InternalTask> taskMap =
                contextReady ? activeTasksByToken : inactiveTasksByToken;
        if (internalTask.isSingleton()) {
            //Reject duplicate singleton tasks - by type + check automatically during insert that we are not
            //rescheduling the same task instance
            if (hasTaskOfType(internalTask.getType(), taskMap, false) ||
                    taskMap.putIfAbsent(token, internalTask) != null) {
                throw new IllegalStateException("Cannot start a singleton task more than once (" + internalTask + ").");
            }
        } else if (taskMap.put(token, internalTask) != null) {
            log.warn("Overriding an active task with the same token {}.", internalTask);
        }
        // Make sure to use the same ready state
        if (contextReady) {
            internalTask.schedule(waitForRunning);
        } else {
            // Recheck may be readiness was activated in between
            if (context.isReady()) {
                moveToActive(token);
            }
        }
        return token;
    }

    @Override
    public String startTask(Task task, boolean waitForRunning, boolean propagateToAffinity) {
        JobName jobName = (JobName) task.getAttribute(Task.JOB_NAME);
        boolean hardAffinity = (boolean) task.getAttribute(Task.HARD_AFFINITY);

        if (propagateToAffinity &&
                !addonsManager.addonByType(HaAddon.class).isAllowedToRun(jobName, hardAffinity)) {
            log.debug("Propagating task {} to Affinity node", task.getToken());
            addonsManager.addonByType(HaAddon.class).propagateTaskToAvailableAffinityNode(task);
            return task.getToken();
        } else {
            return startTask(task, waitForRunning);
        }
    }

    @Override
    public void stopTask(String token, boolean wait) {
        InternalTask task = getActiveInternalTask(activeTasksByToken.get(token));
        if (task != null) {
            task.stop(wait);
        } else {
            log.warn("Could not find task {} to stop.", token);
        }
    }

    @Override
    public void pauseTask(String token, boolean wait) {
        InternalTask task = getActiveInternalTask(token, true);
        if (task != null) {
            task.pause(wait);
        } else {
            log.warn("Could not find task {} to pause.", token);
        }
    }

    /**
     * Checks if task is active, and if so delegates to {@link InternalTask#blockIfPausedAndShouldBreak()} to decide
     * whether pause or stop should be made
     *
     * @return true if should stop, false otherwise
     */
    @Override
    public boolean pauseOrBreak() {
        String token = TaskCallback.currentTaskToken();
        // If not in a task the token is null
        if (token == null) {
            //Since this is called by external clients, it may be called directly or in tests with no surrounding task
            log.debug("No current task is found on thread - nothing to block or pause.");
            return false;
        }
        InternalTask task = getActiveInternalTask(token, true);
        if (task == null) {
            log.warn("Could not find task {} to check block on.", token);
            return false;
        }
        return task.blockIfPausedAndShouldBreak();
    }

    @Override
    public List<String> stopTasks(Class<? extends TaskCallback> callbackType) {
        String currentToken = TaskCallback.currentTaskToken();
        List<String> results = Lists.newArrayList();
        for (Task task : activeTasksByToken.values()) {
            // If callback type is null means all, cannot stop myself
            if ((callbackType == null || ClassUtils.isAssignable(callbackType, task.getType()))
                    && (currentToken == null || !task.getToken().equals(currentToken))) {
                //TODO: Don't wait on each job in a serial fashion
                stopTask(task.getToken(), true);
                results.add(task.getToken());
            }
        }
        return results;
    }

    @Override
    public boolean resumeTask(String token) {
        return resumeTask(token, null);
    }

    @Override
    public boolean resumeTask(String token, String resumingTaskToken) {
        InternalTask task = getActiveInternalTask(token, false);
        if (task != null) {
            return task.resume(resumingTaskToken);
        } else {
            log.debug("Could not find task {} to resume.", token);
            return false;
        }
    }


    @Override
    public void cancelTask(String token, boolean wait) {
        cancelTask(token, wait, ConstantValues.locksTimeoutSecs.getLong());
    }

    @Override
    public void cancelTask(String token, boolean wait, long timeout) {
        InternalTask task = activeTasksByToken.get(token);
        if (task != null) {
            task.cancel(wait, timeout);
        } else {
            log.warn("Could not find task {} to cancel.", token);
        }
        activeTasksByToken.remove(token);
    }

    @Override
    public void cancelTasks(@Nonnull final Class<? extends TaskCallback> callbackType, boolean wait) {
        cancelTasks(TaskUtils.createPredicateForType(callbackType), wait);
    }

    @Override
    public void cancelAllTasks(boolean wait) {
        cancelTasks(input -> true, wait);
    }

    @Override
    public void rescheduleTask(String token) {
        InternalTask task = activeTasksByToken.get(token);
        if (task != null) {
            long waitBeforeExecute =
                    TimeUnit.SECONDS.toMillis(ConstantValues.taskRetryWaitBeforeRescheduleSecs.getInt());
            task.rescheduleTask(waitBeforeExecute);
        } else {
            log.warn("Could not find task {} to reschedule.", token);
        }
    }

    @Override
    public boolean waitForTaskCompletion(String token) {
        InternalTask task = getActiveInternalTask(token, false);
        //Check for null since task may have already been canceled
        return task == null || task.waitForCompletion(0);
    }

    @Override
    public Task getActiveTask(String token, boolean warnIfMissing) {
        if (token == null) {
            throw new IllegalArgumentException("Could not find task with null token");
        }

        Task task = activeTasksByToken.get(token);
        if (warnIfMissing && task == null) {
            log.warn("Could not locate active task with token {}. Task may have been canceled.", token);
        }
        return task;
    }

    InternalTask getActiveInternalTask(Task task) {
        if (task == null) {
            return null;
        }
        return activeTasksByToken.getOrDefault(task.getToken(), (InternalTask) task);
    }

    InternalTask getActiveInternalTask(String token, boolean warnIfMissing) {
        return (InternalTask) getActiveTask(token, warnIfMissing);
    }

    @Override
    public boolean hasTaskOfType(Class<? extends TaskCallback> callbackType) {
        return hasTaskOfType(callbackType, activeTasksByToken, true);
    }

    @Override
    public void cancelTasks(@Nonnull Predicate<Task> predicate, boolean wait) {
        List<Task> toCancel = getActiveTasks(predicate);
        for (Task task : toCancel) {
            //TODO: Don't wait on each job in a serial fashion
            cancelTask(task.getToken(), wait);
        }
    }

    @Override
    public List<Task> getActiveTasks(@Nonnull Predicate<Task> predicate) {
        return activeTasksByToken.values().stream()
                .filter(predicate)
                .collect(Collectors.toList());
    }

    @Override
    public void checkCanStartManualTask(Task task, Class<? extends TaskCallback> typeToRun,
            MutableStatusHolder statusHolder, Object... keyValues) {
        JobCommand jobCommand = typeToRun.getAnnotation(JobCommand.class);
        if (ContextHelper.get().isOffline()) {
            statusHolder.error("Artifactory is in offline state, task are not allowed to run during offline state!",
                    log);
            return;
        }
        if (jobCommand == null) {
            statusHolder.error(
                    "Task type " + typeToRun.getName() + " does not have the " +
                            JobCommand.class.getName() + " annotation!", log);
            return;
        }
        if (jobCommand.manualUser() == TaskUser.INVALID) {
            statusHolder.error("Task type " + typeToRun.getName() + " is not defined to run manually!", log);
            return;
        }
        String currentToken = TaskCallback.currentTaskToken();
        if (currentToken != null) {
            statusHolder
                    .error("Cannot check for manual run with status from inside a running task: " + currentToken, log);
            return;
        }

        // Extract task filters for each stop strategy
        EnumMap<StopStrategy, Predicate<Task>> predicatePerStrategy = getPredicatePerStrategy(typeToRun, keyValues);

        // Check first for impossible to run
        Predicate<Task> impossibleFilter = predicatePerStrategy.get(StopStrategy.IMPOSSIBLE);
        if (impossibleFilter != null) {
            for (InternalTask activeTask : getActiveTasks(impossibleFilter).stream().map(InternalTask.class::cast)
                    .collect(Collectors.toSet())) {
                if (activeTask.processActiveOnCluster()) {
                    statusHolder.error("Task " + typeToRun.getName() + " cannot stop a mandatory related job " +
                            activeTask.getType().getName() + " while it's running!", log);
                }
                if (activeTask.getType().equals(
                        typeToRun) && activeTask.isManuallyActivated() && !activeTask.wasCompleted()) {
                    statusHolder.error("Another manual task " + typeToRun.getName() + " is still active!", log);
                }
            }
        }

        // Then check for stopped
        Predicate<Task> stopFilter = predicatePerStrategy.get(StopStrategy.STOP);
        if (stopFilter != null) {
            for (Task activeTask : getActiveTasks(stopFilter)) {
                if (activeTask.processActiveOnCluster()) {
                    statusHolder.warn(
                            "Task " + activeTask.getType().getName() + " will be stopped by running "
                                    + typeToRun.getName() + " !", log);
                }
            }
        }

        // Then pause what's needed
        Predicate<Task> pauseFilter = predicatePerStrategy.get(StopStrategy.PAUSE);
        if (pauseFilter != null) {
            for (Task activeTask : getActiveTasks(pauseFilter)) {
                if (activeTask.processActiveOnCluster()) {
                    statusHolder.warn(
                            "Task " + activeTask.getType().getName() + " will be paused by running "
                                    + typeToRun.getName() + " !", log);
                }
            }
        }

        InternalTask internalTask = getActiveInternalTask(task);
        if (internalTask.isClusterSingleton()) {
            final String lockKey =
                    internalTask.getTaskLockKeyIncludingKeyAttributes(Arrays.asList(jobCommand.keyAttributes()));
            ConflictGuard jobConflictGuard = getConflictsGuard().getLock(lockKey);
            if (jobConflictGuard.isLocked()) {
                statusHolder.error("Cannot start task (" + task + ") manually!" +
                        " The task lock-key (" + lockKey + ") is already locked.", log);
            }
        }

        if (!statusHolder.isError()) {
            statusHolder.debug("Task " + typeToRun.getName() + " can run.", log);
        }
    }

    @Override
    public void stopRelatedTasks(Class<? extends TaskCallback> typeToRun, List<String> tokenStopped,
            Object... keyValues) {
        // Extract task filters for each stop strategy
        EnumMap<StopStrategy, Predicate<Task>> predicatePerStrategy = getPredicatePerStrategy(typeToRun, keyValues);

        // Each predicates should not count myself :)
        Predicate<Task> notMyself = getNotMySelfPredicate();

        String currentTaskToken = TaskCallback.currentTaskToken();

        // Stop early if impossible to run
        Predicate<Task> impossibleFilter = predicatePerStrategy.get(StopStrategy.IMPOSSIBLE);
        if (impossibleFilter != null) {
            getActiveTasks(notMyself.and(impossibleFilter))
                    .forEach(activeTask -> {
                        InternalTask internalTask = (InternalTask) activeTask;
                        log.debug("Found IMPOSSIBLE task for {}. Task is: {}", typeToRun.getSimpleName(),
                                activeTask.getType());
                        if (internalTask.processActiveOnCluster()) {
                            log.info("IMPOSSIBLE task process is active: {}", activeTask);
                            throw new TaskImpossibleToStartException(
                                    "Job " + typeToRun.getName() + " cannot stop related job " + activeTask +
                                            " while it's running!");
                        }
                        // Immediate stop fails if task in running mode
                        internalTask.stop(false, currentTaskToken);
                        tokenStopped.add(activeTask.getToken());
                    });
        }

        // Then stop what's needed
        Predicate<Task> stopFilter = predicatePerStrategy.get(StopStrategy.STOP);
        if (stopFilter != null) {
            getActiveInternalTasks(notMyself.and(stopFilter))
                    .forEach(activeTask -> {
                        // Just stop and keep for resume
                        log.debug("Found task to STOP for {}. Task token to stop is: {}", typeToRun.getSimpleName(),
                                activeTask.getToken());
                        activeTask.stop(true, currentTaskToken);
                        tokenStopped.add(activeTask.getToken());
                    });
        }

        // Then pause what's needed
        Predicate<Task> pauseFilter = predicatePerStrategy.get(StopStrategy.PAUSE);
        if (pauseFilter != null) {
            getActiveTasks(notMyself.and(pauseFilter)).stream().map(InternalTask.class::cast)
                    .forEach(activeTask -> {
                        // Just stop and always keep for resume, since even single execution paused should be resumed
                        log.debug("Found task to PAUSE for {}. Task token  to pause is: {}", typeToRun.getSimpleName(),
                                activeTask.getToken());
                        activeTask.pause(true, currentTaskToken);
                        tokenStopped.add(activeTask.getToken());
                    });
        }
    }

    @Override
    public boolean waitForTaskCompletion(String token, long timeout) {
        InternalTask task = getActiveInternalTask(token, false);
        //Check for null since task may have already been canceled
        return task == null || task.waitForCompletion(timeout);
    }

    @Override
    public ConflictsGuard<String> getConflictsGuard() {
        return addonsManager.addonByType(HaCommonAddon.class)
                .getConflictsGuard(NonStandardLockCategory.CLUSTER_SINGLETON.getName());
    }

    @Override
    public void canBeStarted(Task task) {
        InternalTask internalTask = getActiveInternalTask(task);
        Class<? extends TaskCallback> callbackType = internalTask.getType();
        JobCommand jobCommand = callbackType.getAnnotation(JobCommand.class);
        if (jobCommand == null) {
            throw new IllegalArgumentException(
                    "Callback type " + callbackType.getName() + " does not have the " +
                            JobCommand.class.getName() + " annotation!");
        }
        if (internalTask.isManuallyActivated() && jobCommand.manualUser() == TaskUser.INVALID) {
            throw new IllegalStateException("Cannot start task (" + internalTask + ") manually!" +
                    " Manual user not defined!");
        }
        boolean hasKeyAttributes = jobCommand.keyAttributes().length > 0;
        if (!jobCommand.singleton() && !hasKeyAttributes) {
            return;
        }
        EnumMap<StopStrategy, Predicate<Task>> predicatePerStrategy = getPredicatePerStrategy(callbackType,
                internalTask.getKeyValues());
        // Should not count myself :)
        Predicate<Task> notMyself = getNotMySelfPredicate();

        // Only impossible are blocker here
        Predicate<Task> impossibleFilter = predicatePerStrategy.get(StopStrategy.IMPOSSIBLE);
        if (impossibleFilter != null) {
            for (InternalTask activeTask : getActiveTasks(notMyself.and(impossibleFilter)).stream()
                    .map(InternalTask.class::cast).collect(Collectors.toSet())) {
                if (internalTask.isManuallyActivated()) {
                    if (activeTask.processActiveOnCluster()) {
                        throw new IllegalStateException("Cannot start task (" + internalTask + ") manually!" +
                                " Task " + activeTask + " is active!");
                    }
                    if (activeTask.isManuallyActivated() && !activeTask.wasCompleted()) {
                        throw new IllegalStateException("Cannot start task (" + internalTask + ") manually!" +
                                " Another manual task " + activeTask + " exists and is not completed!");
                    }
                } else {
                    // Starting impossible filter of same type means singleton or key based singleton match
                    if (activeTask.getType().equals(callbackType)) {
                        if (activeTask.isManuallyActivated()) {
                            if (!activeTask.wasCompleted()) {
                                throw new IllegalStateException("Cannot schedule task (" + internalTask + ")!" +
                                        " Another manual task of same type and key " + activeTask +
                                        " exists and is not completed!");
                            }
                        } else {
                            throw new IllegalStateException("Cannot schedule task (" + internalTask + ")!" +
                                    " Another task of same type and key " + activeTask + " is already scheduled!");
                        }
                    }
                }
            }
        }
    }

    /**
     * Used only by the task in case that the task needs to cancel itself
     */
    @Override
    public void cancelMyselfCallback(Task task) {
        activeTasksByToken.remove(task.getToken());
        ((InternalTask) task).cancelTask();
    }

    private void activateTasks() {
        Iterator<String> taskTokens = inactiveTasksByToken.keySet().iterator();
        while (taskTokens.hasNext()) {
            String taskKey = taskTokens.next();
            //Check the readiness and task status again since it can change
            if (ContextHelper.get().isReady()) {
                moveToActive(taskKey);
                taskTokens.remove();
            }
        }
    }

    private void moveToActive(String taskKey) {
        InternalTask task = inactiveTasksByToken.get(taskKey);
        if (task != null) {
            activeTasksByToken.put(taskKey, task);
            task.schedule(false);
        }
    }

    private Predicate<Task> getNotMySelfPredicate() {
        return input -> {
            // null task always false
            if (input == null) {
                return false;
            }
            String currentToken = TaskCallback.currentTaskToken();
            return currentToken == null || !currentToken.equals(input.getToken());
        };
    }

    private EnumMap<StopStrategy, Predicate<Task>> getPredicatePerStrategy(
            Class<? extends TaskCallback> typeToRun,
            final Object... keyValues) {
        EnumMap<StopStrategy, Predicate<Task>> result = new EnumMap<>(StopStrategy.class);
        JobCommand toRunJobCommand = typeToRun.getAnnotation(JobCommand.class);
        final String[] keys = toRunJobCommand.keyAttributes();
        if (keys.length != keyValues.length) {
            throw new IllegalArgumentException(
                    "Cannot check if task " + typeToRun + " can start without key values!\n" +
                            "Received " + Arrays.toString(keyValues) + " and expected values for " + Arrays.toString(
                            keys));
        }
        for (StopCommand stopCommand : toRunJobCommand.commandsToStop()) {
            Class<? extends TaskCallback> callbackType;
            // get callback type from command or command name
            callbackType = getCommand(typeToRun, stopCommand);
            if ((callbackType.getModifiers() & Modifier.ABSTRACT) != 0) {
                throw new IllegalArgumentException(
                        "Job command definition for " + typeToRun.getName() + " contain an abstract class to stop!");
            }
            StopStrategy stopStrategy = stopCommand.strategy();
            TaskTypePredicate taskPredicate = (TaskTypePredicate) result.get(stopStrategy);
            if (stopCommand.useKey()) {
                taskPredicate = getTaskPredicateForKey(callbackType, taskPredicate, keyValues);
            } else {
                taskPredicate = getTaskPredicateForType(callbackType, taskPredicate);
            }
            result.put(stopStrategy, taskPredicate);
        }

        if (toRunJobCommand.singleton()) {
            // If singleton and currently running => Impossible to stop
            TaskTypePredicate taskPredicate = (TaskTypePredicate) result.get(StopStrategy.IMPOSSIBLE);
            taskPredicate = getTaskPredicateForType(typeToRun, taskPredicate);
            result.put(StopStrategy.IMPOSSIBLE, taskPredicate);
        }

        if (keys.length != 0) {
            // If key based and currently running => Impossible to stop
            TaskTypePredicate taskPredicate = (TaskTypePredicate) result.get(StopStrategy.IMPOSSIBLE);
            taskPredicate = getTaskPredicateForKey(typeToRun, taskPredicate, keyValues);
            result.put(StopStrategy.IMPOSSIBLE, taskPredicate);
        }

        return result;
    }

    /**
     * get task to do stop command on it
     *
     * @param stopCommand - stop command to execute on job
     * @return callback task class
     */
    private Class<? extends TaskCallback> getCommand(Class<? extends TaskCallback> typeToRun, StopCommand stopCommand) {

        Class<? extends TaskCallback> callbackType = null;
        ScheduleJobEnum commandName = stopCommand.commandName();
        // is command name define bu no command type
        if (!commandName.equals(ScheduleJobEnum.DUMMY_JOB) && stopCommand.command() == DummyJob.class) {
            try {
                callbackType = (Class<? extends TaskCallback>) Class.forName(commandName.jobName);
            } catch (ClassNotFoundException e) {
                log.error("Class commandName not found", e);
            }
        }
        // no command name or command type has been define
        else if (commandName.equals(ScheduleJobEnum.DUMMY_JOB) && stopCommand.command() == DummyJob.class) {
            log.error("No job command type or name was defined for job class: {}", typeToRun.getSimpleName());
        } else {//command type has been define
            callbackType = stopCommand.command();
        }
        return callbackType;
    }

    private TaskTypePredicate getTaskPredicateForType(Class<? extends TaskCallback> callbackType,
            TaskTypePredicate taskPredicate) {
        if (taskPredicate == null) {
            taskPredicate = new TaskTypePredicate(callbackType);
        } else {
            taskPredicate = new TaskTypePredicate(taskPredicate, callbackType);
        }
        return taskPredicate;
    }

    private TaskTypePredicate getTaskPredicateForKey(final Class<? extends TaskCallback> callbackType,
            TaskTypePredicate taskPredicate, final Object[] keyValues) {
        if (taskPredicate == null) {
            taskPredicate = new TaskTypePredicate();
        }
        taskPredicate.addSubPredicate(input -> (input != null) &&
                ClassUtils.isAssignable(callbackType, input.getType()) &&
                input.keyEquals(keyValues));
        return taskPredicate;
    }

    private boolean hasTaskOfType(Class<? extends TaskCallback> callbackType,
            ConcurrentMap<String, InternalTask> taskMap,
            boolean withManual) {
        if (callbackType == null) {
            return false;
        }
        for (InternalTask task : taskMap.values().stream().map(InternalTask.class::cast).collect(Collectors.toSet())) {
            if (ClassUtils.isAssignable(callbackType, task.getType()) &&
                    !(withManual && task.isManuallyActivated())) {
                return true;
            }
        }
        return false;
    }

    private List<InternalTask> getActiveInternalTasks(@Nonnull Predicate<Task> predicate) {
        return getActiveTasks(predicate).stream().map(InternalTask.class::cast).collect(Collectors.toList());
    }
}
