package org.artifactory.servicetrust;

import com.google.common.annotations.VisibleForTesting;
import org.apache.http.HttpEntity;
import org.apache.http.HttpHeaders;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.util.EntityUtils;
import org.artifactory.api.security.access.CreatedTokenInfo;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.model.xstream.security.UserImpl;
import org.artifactory.security.SimpleUser;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.spring.InternalContextHelper;
import org.artifactory.util.HttpClientConfigurator;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.common.SubjectFQN;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.access.token.JwtAccessTokenImpl;
import org.jfrog.common.JsonUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import static org.artifactory.servicetrust.TokenSpecService.INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX;
import static org.artifactory.servicetrust.TokenSpecService.INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX;
import static org.artifactory.servicetrust.TokenSpecService.SCOPE_PATTERN;

/**
 * @author igoru
 */
@Service
public class ServiceTrustServiceImpl implements ServiceTrustService {

    private static final int SOCKET_TIMEOUT = 15000;
    private static final int CONNECTION_TIMEOUT = 1500;
    private static final String BEARER = "Bearer ";
    private static final String ERROR_EXCHANGING_TOKEN = "Error exchanging token %d, %s";
    private static final Set<ServiceTrustUseCase> USE_CASES
            = Set.of(ServiceTrustUseCase.ARTIFACTORY_COLD, ServiceTrustUseCase.ARTIFACTORY_COLD_WITH_NAMESPACE);
    private static final String USE_CASE_NOT_FOUND = "Use case %s not found";

    private static final Map<ServiceTrustUseCase, TokenSpecService> SPEC_SERVICES;

    static {
        SPEC_SERVICES = new HashMap<>();
        SPEC_SERVICES.put(ServiceTrustUseCase.ARTIFACTORY_COLD,
                ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE);
        SPEC_SERVICES
                .put(ServiceTrustUseCase.ARTIFACTORY_COLD_WITH_NAMESPACE,
                        ArtifactoryColdTokenSpecService.ARTIFACTORY_COLD_TOKEN_SPEC_SERVICE);
    }

    @Autowired
    private AccessService accessService;

    @Override
    public String getPairingToken(String useCasePath, ServiceTrust trust, String baseUrl, String pairingUrl,
            String exchangeUrl) {
        TokenSpec pairingTokenSpec = createPairingTokenSpec(useCasePath, trust, baseUrl, pairingUrl, exchangeUrl);
        CreatedTokenInfo tokenInfo = accessService.createToken(pairingTokenSpec);
        return tokenInfo.getTokenValue();
    }

    @Override
    public void pairToken(String useCasePath, ServiceTrustUseCase serviceTrustUseCase, String token)
            throws IOException {
        JwtAccessToken jwt = parseTokenValue(token);
        TokenExtension extension = JsonUtils.getInstance().readValue(jwt.getExtension(), TokenExtension.class);
        ServiceId artifactoryServiceId = accessService.getArtifactoryServiceId();
        String subject = createTokenSubject(artifactoryServiceId);
        ExchangeTokenRequest exchangeTokenRequest =
                getSpecService(serviceTrustUseCase).createExchangeTokenRequest(subject, useCasePath, jwt);
        ExchangeTokenResponse exchangeTokenResponse =
                executeTokenExchangeRequest(extension.getExchangeUrl(), token, exchangeTokenRequest);
        storeMasterTokenData(serviceTrustUseCase, exchangeTokenResponse, extension);
    }

    @Override
    public void refreshMasterToken(ServiceTrustUseCase serviceTrustUseCase, String exchangeUrl, String masterToken)
            throws IOException {
        ServiceId artifactoryServiceId = accessService.getArtifactoryServiceId();
        String subject = createTokenSubject(artifactoryServiceId);
        ExchangeTokenRequest exchangeTokenRequest =
                getSpecService(serviceTrustUseCase).createMasterTokenExchangeRequest(subject);
        ExchangeTokenResponse exchangeTokenResponse =
                executeTokenExchangeRequest(exchangeUrl, masterToken, exchangeTokenRequest);
        updateMasterTokenData(serviceTrustUseCase, exchangeTokenResponse);
    }

    @Override
    public ExchangeTokenResponse exchangeMasterToken(ServiceTrustUseCase serviceTrustUseCase,
            String exchangeUrl, String masterToken) throws IOException {
        ServiceId artifactoryServiceId = accessService.getArtifactoryServiceId();
        String subject = createTokenSubject(artifactoryServiceId);
        ExchangeTokenRequest exchangeTokenRequest =
                getSpecService(serviceTrustUseCase).createScopedTokenExchangeRequest(subject);
        return executeTokenExchangeRequest(exchangeUrl, masterToken, exchangeTokenRequest);
    }

    @Override
    public ExchangeTokenResponse exchangeToken(ExchangeTokenRequest request) {
        Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
        AccessTokenAuthentication accessTokenAuthentication = (AccessTokenAuthentication) authentication;
        JwtAccessToken authToken = accessTokenAuthentication.getAccessToken();
        List<String> scopes = authToken.getScope();
        String tokenValue = authToken.getTokenValue();
        if (tokenValue == null) {
            throw new ServiceTrustException("Token value is null");
        }
        if (scopes.size() == 1) {
            return pairing(request, accessTokenAuthentication, scopes, tokenValue);
        } else {
            String pairingScope = findPairingScope(scopes);
            String useCasePath = getUseCasePathFromScope(pairingScope);
            String target = request.getTarget();
            if (target.equals(useCasePath)) {
                return createNewMaster(request, accessTokenAuthentication, tokenValue);
            } else {
                return exchange(accessTokenAuthentication, target);
            }
        }
    }

    @Override
    public Optional<ServiceTrustUseCase> findUseCase(String useCasePath) {
        return USE_CASES.stream().filter(u -> u.matches(useCasePath)).findFirst();
    }

    @VisibleForTesting
    void storeMasterTokenData(ServiceTrustUseCase useCase, ExchangeTokenResponse response, TokenExtension extension) {
        getSpecService(useCase).storeMasterTokenData(response, extension);
    }

    @VisibleForTesting
    private void updateMasterTokenData(ServiceTrustUseCase serviceTrustUseCase,
            ExchangeTokenResponse exchangeTokenResponse) {
        getSpecService(serviceTrustUseCase).updateMasterTokenData(exchangeTokenResponse);
    }

    @VisibleForTesting
    JwtAccessToken parseTokenValue(String token) {
        return JwtAccessTokenImpl.parseTokenValue(token);
    }

    @VisibleForTesting
    ExchangeTokenResponse createToken(AccessTokenAuthentication accessTokenAuthentication, TokenSpec tokenSpec) {
        elevateToAdmin(accessTokenAuthentication);
        CreatedTokenInfo masterTokenInfo = accessService.createToken(tokenSpec, true);
        String masterToken = masterTokenInfo.getTokenValue();
        JwtAccessToken jwt = parseTokenValue(masterToken);
        return new ExchangeTokenResponse(jwt.getTokenId(), masterToken, masterTokenInfo.getExpiresIn(),
                masterTokenInfo.getScope());
    }

    @VisibleForTesting
    CloseableHttpResponse executeMethod(HttpPost postMethod) throws IOException {
        ProxyDescriptor proxy = InternalContextHelper.get().getCentralConfig().getDescriptor().getDefaultProxy();
        return new HttpClientConfigurator()
                .proxy(proxy)
                .socketTimeout(SOCKET_TIMEOUT)
                .connectionTimeout(CONNECTION_TIMEOUT)
                .noRetry()
                .build()
                .execute(postMethod);
    }

    @VisibleForTesting
    void setAccessService(AccessService accessService) {
        this.accessService = accessService;
    }

    private ExchangeTokenResponse executeTokenExchangeRequest(String exchangeUrl, String token,
            ExchangeTokenRequest exchangeTokenRequest) throws IOException {
        String requestBody = JsonUtils.getInstance().valueToString(exchangeTokenRequest, true);
        HttpEntity requestEntity = new StringEntity(requestBody, ContentType.APPLICATION_JSON);
        HttpPost postMethod = new HttpPost(exchangeUrl);
        postMethod.setEntity(requestEntity);
        postMethod.addHeader(HttpHeaders.AUTHORIZATION, BEARER + token);
        try (CloseableHttpResponse response = executeMethod(postMethod)) {
            int statusCode = response.getStatusLine().getStatusCode();
            if (statusCode != HttpStatus.SC_OK) {
                String error = EntityUtils.toString(response.getEntity());
                String errorMessage = String.format(ERROR_EXCHANGING_TOKEN, statusCode, error);
                throw new ServiceTrustException(errorMessage);
            }
            HttpEntity responseEntity = response.getEntity();
            return JsonUtils.getInstance().readValue(responseEntity.getContent(), ExchangeTokenResponse.class);
        }
    }

    private TokenSpecService getSpecService(ServiceTrustUseCase useCase) {
        TokenSpecService factory = SPEC_SERVICES.get(useCase);
        if (factory == null) {
            throw new ServiceTrustException(String.format("No spec service for usecase %s", useCase));
        }
        return factory;
    }

    private TokenSpec createExchangeTokenSpec(String useCasePath) {
        ServiceTrustUseCase useCase = findUseCase(useCasePath)
                .orElseThrow(() -> new ServiceTrustException(String.format(USE_CASE_NOT_FOUND, useCasePath)));
        ServiceId artifactoryServiceId = accessService.getArtifactoryServiceId();
        TokenSpecService specService = getSpecService(useCase);
        return specService.createExchangeTokenSpec(artifactoryServiceId, useCasePath);
    }

    private TokenSpec createMasterTokenSpec(String target, AccessTokenAuthentication accessTokenAuthentication) {
        ServiceTrustUseCase useCase = findUseCase(target)
                .orElseThrow(() -> new ServiceTrustException(String.format(USE_CASE_NOT_FOUND, target)));
        ServiceId artifactoryServiceId = accessService.getArtifactoryServiceId();
        JwtAccessToken token = accessTokenAuthentication.getAccessToken();
        String pairingScope = findPairingScope(token.getScope());
        TokenExtension extension = JsonUtils.getInstance().readValue(token.getExtension(), TokenExtension.class);
        TokenSpecService specService = getSpecService(useCase);
        return specService.createMasterTokenSpec(artifactoryServiceId, target, pairingScope, extension.getBaseUrl(),
                extension.getExchangeUrl());
    }

    private TokenSpec createPairingTokenSpec(String useCasePath, ServiceTrust trust, String baseUrl, String pairingUrl,
            String exchangeUrl) {
        ServiceTrustUseCase useCase = findUseCase(useCasePath)
                .orElseThrow(() -> new ServiceTrustException(String.format(USE_CASE_NOT_FOUND, useCasePath)));
        ServiceId artifactoryServiceId = accessService.getArtifactoryServiceId();
        String subject = createTokenSubject(artifactoryServiceId);
        String scope = String.format(SCOPE_PATTERN, useCasePath);
        TokenSpecService specService = getSpecService(useCase);
        return specService
                .createPairingTokenSpec(artifactoryServiceId, trust, subject, scope, baseUrl, pairingUrl, exchangeUrl);
    }

    private String createTokenSubject(ServiceId artifactoryServiceId) {
        Authentication auth = SecurityContextHolder.getContext().getAuthentication();
        if (auth instanceof AccessTokenAuthentication) {
            AccessTokenAuthentication a = (AccessTokenAuthentication) auth;
            JwtAccessToken accessToken = a.getAccessToken();
            return accessToken.getSubject();
        } else {
            return createSubjectForUsername(artifactoryServiceId, auth);
        }
    }

    private String createSubjectForUsername(ServiceId artifactoryServiceId, Authentication auth) {
        String username = auth.getName();
        SubjectFQN subjectFQN = new SubjectFQN(artifactoryServiceId, SubjectFQN.USERS_NAME_PART, username);
        return subjectFQN.getFormattedName();
    }

    private ExchangeTokenResponse exchange(AccessTokenAuthentication accessTokenAuthentication,
            String useCasePath) {
        TokenSpec exchangeTokenSpec = createExchangeTokenSpec(useCasePath);
        return createToken(accessTokenAuthentication, exchangeTokenSpec);
    }

    private ExchangeTokenResponse createNewMaster(ExchangeTokenRequest request,
            AccessTokenAuthentication accessTokenAuthentication, String tokenValue) {
        TokenSpec masterTokenSpec = createMasterTokenSpec(request.getTarget(), accessTokenAuthentication);
        accessService.revokeToken(tokenValue);
        return createToken(accessTokenAuthentication, masterTokenSpec);
    }

    private ExchangeTokenResponse pairing(ExchangeTokenRequest request,
            AccessTokenAuthentication accessTokenAuthentication, List<String> scopes, String tokenValue) {
        String pairingScope = findPairingScope(scopes);
        String useCasePath = getUseCasePathFromScope(pairingScope);
        if (!request.getTarget().equals(useCasePath)) {
            throw new ServiceTrustException("Target and usecase doesnt match");
        }
        return createNewMaster(request, accessTokenAuthentication, tokenValue);
    }

    private String getUseCasePathFromScope(String pairingScope) {
        return pairingScope.replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX, "")
                .replace(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX, "");
    }

    private void elevateToAdmin(AccessTokenAuthentication accessTokenAuthentication) {
        JwtAccessToken authToken = accessTokenAuthentication.getAccessToken();
        SimpleUser principal = (SimpleUser) accessTokenAuthentication.getPrincipal();
        UserImpl descriptor = new UserImpl(principal.getDescriptor());
        descriptor.setAdmin(true);
        SimpleUser elevatedPrincipal = new SimpleUser(descriptor);
        AccessTokenAuthentication elevated = new AccessTokenAuthentication(authToken,
                elevatedPrincipal, accessTokenAuthentication.getAuthorities());
        elevated.setAuthenticated(true);
        SecurityContextHolder.getContext().setAuthentication(elevated);
    }

    private String findPairingScope(List<String> scopes) {
        return scopes.stream()
                .filter(s -> s.startsWith(INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX))
                .findFirst()
                .orElseThrow(() -> new ServiceTrustException("Pairing scope not found"));
    }
}
