package org.artifactory.metrics.providers.features;

import ch.qos.logback.classic.Level;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Lists;
import com.jfrog.test.LogTestUtils;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.artifactory.api.callhome.FeatureGroup;
import org.artifactory.api.storage.BinariesInfo;
import org.artifactory.storage.StorageService;
import org.artifactory.storage.StorageSummaryInfo;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersInfo;
import org.jfrog.storage.binstore.ifc.model.BinaryTreeElement;
import org.mockito.Mock;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Collections;
import java.util.Map;

import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

public class StorageFeatureTest {

    private static final Logger log = LoggerFactory.getLogger(StorageFeatureTest.class);

    @Mock
    private StorageService storageService;
    @Mock
    private StorageSummaryInfo storageSummaryInfo;
    @Mock
    private BinariesInfo binariesInfo;
    @Mock
    private Map<String, Object> nextData;
    @Mock
    private Map<String, Object> nextSubData;
    @Mock
    private BinaryTreeElement rootTreeElement;
    @Mock
    private BinaryTreeElement nextTreeElement;
    @Mock
    private BinaryTreeElement nextSubElement;

    private ObjectMapper objectMapper = new ObjectMapper();

    @BeforeClass
    public void setLog() {
        LogTestUtils.setLoggingLevel(StorageFeatureTest.class, Level.INFO);
    }

    @BeforeMethod
    public void setup() {
        openMocks(this);

        when(rootTreeElement.getData()).thenReturn(Collections.singletonMap("type", "federated-repo"));

        when(rootTreeElement.getNextBinaryTreeElement()).thenReturn(nextTreeElement);
        when(nextTreeElement.getData()).thenReturn(nextData);

        when(nextData.get(eq("type"))).thenReturn("file-system");
        when(nextData.get(eq("id"))).thenReturn("file-system");
        when(nextData.get(eq("usageSpace"))).thenReturn(String.valueOf(Long.MAX_VALUE));
        when(nextData.get(eq("freeSpace"))).thenReturn(String.valueOf(Long.MAX_VALUE));
        when(nextData.get(eq("totalSpace"))).thenReturn(String.valueOf(Long.MAX_VALUE));

        BinaryProvidersInfo binaryProvidersInfo = new BinaryProvidersInfo(null, rootTreeElement);
        when(storageService.getBinaryProviderInfo()).thenReturn(binaryProvidersInfo);
        when(storageSummaryInfo.getBinariesInfo()).thenReturn(binariesInfo);
        when(storageService.getStorageSummaryInfo()).thenReturn(storageSummaryInfo);

        when(nextSubElement.getData()).thenReturn(nextSubData);
        when(nextSubData.get(eq("type"))).thenReturn("sub-file-system");
        when(nextSubData.get(eq("id"))).thenReturn("sub-file-system");
        when(nextSubData.get(eq("usageSpace"))).thenReturn(String.valueOf(Long.MAX_VALUE));
        when(nextSubData.get(eq("freeSpace"))).thenReturn(String.valueOf(Long.MAX_VALUE));
        when(nextSubData.get(eq("totalSpace"))).thenReturn(String.valueOf(Long.MAX_VALUE));
    }

    @Test
    public void getFeatureFederatedRepoBinaryTreeElementNoMirrors() throws Exception {
        StorageFeature storageFeature = new StorageFeature();
        FieldUtils.writeField(storageFeature, "storageService", storageService, true);
        FeatureGroup featureGroup = storageFeature.getFeature();

        verify(featureGroup, "file-system", "file-system");
    }

    @Test
    public void getFeatureFederatedRepoBinaryTreeElementWithMirrors() throws Exception {
        when(nextTreeElement.getSubBinaryTreeElements()).thenReturn(Lists.newArrayList(nextSubElement));

        StorageFeature storageFeature = new StorageFeature();
        FieldUtils.writeField(storageFeature, "storageService", storageService, true);
        FeatureGroup featureGroup = storageFeature.getFeature();

        verify(featureGroup, "file-system", "sub-file-system");
    }

    private void verify(FeatureGroup featureGroup, String root, String sub) throws Exception {
        String json = objectMapper.writerWithDefaultPrettyPrinter().writeValueAsString(featureGroup);
        log.info(json);

        JsonNode node = objectMapper.readTree(json);
        JsonNode features = node.get("features");
        Assert.assertEquals(features.get(0).get("attributes").get("provider").asText(), root);
        Assert.assertEquals(features.get(1).get("name").asText(), "mount " + sub);
        Assert.assertEquals(features.get(1).get("attributes").get("label").asText(), sub);
        Assert.assertEquals(features.get(1).get("attributes").get("type").asText(), sub);
    }
}
