package org.artifactory.servicetrust;

import org.apache.http.HttpEntity;
import org.apache.http.HttpHeaders;
import org.apache.http.HttpStatus;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.util.EntityUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.access.client.token.TokenVerifyResult;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.common.JsonUtils;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.Arrays;

import static org.artifactory.servicetrust.TokenSpecService.INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX;
import static org.artifactory.servicetrust.TokenSpecService.INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertEquals;

public class ServiceTrustPairingTest extends ArtifactoryHomeBoundTest {

    AccessService accessService;

    @BeforeMethod
    private void beforeMethod() {
        InternalArtifactoryContext context = mock(InternalArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        ArtifactRetentionAddon artifactRetentionAddon = mock(ArtifactRetentionAddon.class);
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);
        AddonsManager addonsManager = mock(AddonsManager.class);
        when(addonsManager.addonByType(ArtifactRetentionAddon.class)).thenReturn(artifactRetentionAddon);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
    }

    @AfterMethod
    public void afterMethod() {
        ArtifactoryContextThreadBinder.unbind();
        SecurityContextHolder.clearContext();
    }

    @Test
    public void happyFlow() throws IOException {
        ServiceTrustServiceForTest service = createService();
        setupUserAuthentication();
        TokenVerifyResult verifyResult = mock(TokenVerifyResult.class);
        when(verifyResult.isSuccessful()).thenReturn(true);
        when(accessService.verifyAndGetResult(any())).thenReturn(verifyResult);

        service.pairToken("artifactory-cold", ServiceTrustUseCase.ARTIFACTORY_COLD, "token");

        assertEquals(service.storedUseCase, ServiceTrustUseCase.ARTIFACTORY_COLD);
        assertEquals(service.storedResponse.getAccessToken(), "accessToken");
        assertEquals(service.storedResponse.getAccessToken(), "accessToken");
        assertEquals(service.storedResponse.getExpiresIn(), Long.valueOf(42));
        assertEquals(service.storedResponse.getScope(), "scope");
        assertEquals(service.storedResponse.getTokenType(), "access_token");
    }

    @Test(expectedExceptions = ServiceTrustException.class, expectedExceptionsMessageRegExp =
            "Error exchanging token" + " 500, server error")
    public void whenExchangeFailedThenExceptionThrown() throws IOException {
        ServiceTrustServiceForTest service = createService();
        service.failOnExchange = true;
        setupUserAuthentication();
        TokenVerifyResult verifyResult = mock(TokenVerifyResult.class);
        when(verifyResult.isSuccessful()).thenReturn(true);
        when(accessService.verifyAndGetResult(any())).thenReturn(verifyResult);

        service.pairToken("artifactory-cold", ServiceTrustUseCase.ARTIFACTORY_COLD, "token");
    }

    @Test
    public void testExchangeRequestParameters() throws IOException {
        ServiceTrustServiceForTest service = createService();
        setupUserAuthentication();
        TokenVerifyResult verifyResult = mock(TokenVerifyResult.class);
        when(verifyResult.isSuccessful()).thenReturn(true);
        when(accessService.verifyAndGetResult(any())).thenReturn(verifyResult);

        service.pairToken("artifactory-cold", ServiceTrustUseCase.ARTIFACTORY_COLD, "token");

        assertEquals(service.executedMethod.getURI().toString(), "exchangeUrl");
        String entity = EntityUtils.toString(service.executedMethod.getEntity());
        ExchangeTokenRequest request = JsonUtils.getInstance().readValue(entity, ExchangeTokenRequest.class);
        assertEquals(request.getSubject(), "a@b/users/username");
        assertEquals(request.getTarget(), "artifactory-cold");
        assertEquals(Arrays.stream(service.executedMethod.getHeaders(HttpHeaders.AUTHORIZATION)).iterator().next()
                        .getValue(),
                "Bearer token");
    }

    private ServiceTrustServiceForTest createService() {
        ServiceId serviceId = new ServiceId("a", "b");
        ServiceTrustServiceForTest service = new ServiceTrustServiceForTest();
        accessService = mock(AccessService.class);
        when(accessService.getArtifactoryServiceId()).thenReturn(serviceId);
        service.setAccessService(accessService);
        return service;
    }

    private void setupUserAuthentication() {
        Authentication authentication = mock(Authentication.class);
        when(authentication.getName()).thenReturn("username");
        SecurityContext securityContext = mock(SecurityContext.class);
        SecurityContextHolder.setContext(securityContext);
        when(securityContext.getAuthentication()).thenReturn(authentication);
    }

    private static class ServiceTrustServiceForTest extends ServiceTrustServiceImpl {

        private ServiceTrustUseCase storedUseCase;
        private ExchangeTokenResponse storedResponse;
        private HttpPost executedMethod;
        private boolean failOnExchange;

        @Override
        protected JwtAccessToken parseTokenValue(String token) {
            TokenExtension extension = new TokenExtension("baseUrl", "exchangeUrl", "pairingUrl", false);
            JwtAccessToken jwt = mock(JwtAccessToken.class);
            when(jwt.getExtension()).thenReturn(JsonUtils.getInstance().valueToString(extension));
            String scope = String.join("", INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_PREFIX, "artifactory-cold",
                    INTERNAL_SERVICE_TRUST_PAIRING_SCOPE_POSTFIX);
            when(jwt.getScope()).thenReturn(Arrays.asList(scope));
            return jwt;
        }

        @Override
        CloseableHttpResponse executeMethod(HttpPost postMethod) {
            if (failOnExchange) {
                StatusLine statusLine = mock(StatusLine.class);
                when(statusLine.getStatusCode()).thenReturn(HttpStatus.SC_INTERNAL_SERVER_ERROR);
                HttpEntity responseEntity = new StringEntity("server error", ContentType.APPLICATION_JSON);
                CloseableHttpResponse response = mock(CloseableHttpResponse.class);
                when(response.getStatusLine()).thenReturn(statusLine);
                when(response.getEntity()).thenReturn(responseEntity);
                return response;
            }
            executedMethod = postMethod;
            StatusLine statusLine = mock(StatusLine.class);
            when(statusLine.getStatusCode()).thenReturn(HttpStatus.SC_OK);
            ExchangeTokenResponse exchangeTokenResponse =
                    new ExchangeTokenResponse("tokenId", "accessToken", 42L, "scope");
            String exchangeTokenResponseJson = JsonUtils.getInstance().valueToString(exchangeTokenResponse);
            HttpEntity responseEntity = new StringEntity(exchangeTokenResponseJson, ContentType.APPLICATION_JSON);
            CloseableHttpResponse response = mock(CloseableHttpResponse.class);
            when(response.getStatusLine()).thenReturn(statusLine);
            when(response.getEntity()).thenReturn(responseEntity);
            return response;
        }

        @Override
        void storeMasterTokenData(ServiceTrustUseCase useCase, ExchangeTokenResponse response,
                TokenExtension extension) {
            storedUseCase = useCase;
            storedResponse = response;
        }
    }
}
