package org.artifactory.lifecycle.storage.db.init;

import lombok.SneakyThrows;
import org.artifactory.common.ha.HaNodeProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.storage.db.dao.locks.LockInfo;
import org.artifactory.storage.db.locks.service.DbLocksService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoSession;
import org.mockito.quality.Strictness;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.Supplier;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;


public class HaInitLockTest extends ArtifactoryHomeBoundTest {

    private static final String SERVER_ID = "serverId";

    private MockitoSession mockitoSession;

    private final File releaseInitMarkerFile;

    private HaInitLock haInitLock;
    @Mock
    private DbLocksService dbLocksService;
    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private HaNodeProperties haNodeProperties;
    @Mock
    private Runnable task;
    @Mock
    private Supplier<Boolean> shouldLockSupplier;

    public HaInitLockTest() throws IOException {
        releaseInitMarkerFile = File.createTempFile(".release.init.marker", "");
    }

    @BeforeMethod
    public void setUp() {
        mockitoSession = Mockito.mockitoSession()
                .initMocks(this)
                .strictness(Strictness.STRICT_STUBS)
                .startMocking();
        haInitLock = new HaInitLock(dbLocksService, artifactoryHome);
    }

    @AfterMethod
    public void tearDown() {
        mockitoSession.finishMocking();
    }

    @Test
    public void haEnvMarkerExists() {
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);
        givenReleaseInitLockMarkerFileExists();

        haInitLock.runInsideInitLockIfNeeded(task, () -> true);

        verifyForceUnlockInvoked();
        verifyReleaseInitLockMarkerFileDeleted();
        verifyAcquireLockInvoked();
        verifyTaskRunInvoked(1);
    }

    @Test
    public void haEnvMarkerDoesntExistDbUnLocked() {
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);
        givenReleaseInitLockMarkerFileDoesntExist();
        givenDbUnlocked();

        haInitLock.runInsideInitLockIfNeeded(task, () -> true);

        verifyForceUnlockInvoked(0);
        verifyAcquireLockInvoked();
        verifyTaskRunInvoked(1);
    }

    @Test
    public void haEnvReleaseInitMarkerDoesntExistDbLockedByOwner() {
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);
        givenReleaseInitLockMarkerFileDoesntExist();
        givenDbLockedByOwner();

        haInitLock.runInsideInitLockIfNeeded(task, () -> true);

        verifyForceUnlockInvoked();
        verifyAcquireLockInvoked();
        verifyTaskRunInvoked(1);
    }

    @Test
    public void haEnvReleaseInitMarkerDoesntExistDbLockedByOther() {
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);
        givenReleaseInitLockMarkerFileDoesntExist();
        givenDbLockedByOther();

        haInitLock.runInsideInitLockIfNeeded(task, () -> true);

        verifyForceUnlockInvoked(0);
        verifyAcquireLockInvoked();
        verifyTaskRunInvoked(1);

    }

    @Test(expectedExceptions = IllegalStateException.class, expectedExceptionsMessageRegExp = ".*Unable to acquire " +
            "high-availability lock.*")
    public void haEnvAcquireLockThrowsException() {
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);
        givenDbUnlocked();
        givenReleaseInitLockMarkerFileDoesntExist();
        givenAcquireLockThrowsException();

        haInitLock.runInsideInitLockIfNeeded(task, () -> true);
    }

    @Test
    public void testInitLockAcquiredAndReleased() {
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        when(shouldLockSupplier.get()).thenReturn(true, false);
        givenReleaseInitLockMarkerFileDoesntExist();
        givenDbUnlocked();
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(haNodeProperties.getServerId()).thenReturn("serverId");

        haInitLock.runInsideInitLockIfNeeded(task, shouldLockSupplier);

        verifyAcquireLockInvoked();
        verify(dbLocksService, times(1)).unlock(NonStandardLockCategory.CLUSTER_SINGLETON.getName(),
                HaInitLock.ARTIFACTORY_INIT_LOCK_KEY, "serverId");
        verifyTaskRunInvoked(1);
    }

    @Test
    public void testInitLockAcquiredAndNotReleased() {
        givenReleaseInitLockMarkerFileDoesntExist();
        givenDbUnlocked();
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);

        haInitLock.runInsideInitLockIfNeeded(task, () -> true);

        verifyAcquireLockInvoked();
        verify(dbLocksService, times(0)).forceUnlock(any(), any());
        verifyNoMoreInteractions(dbLocksService);
        verifyTaskRunInvoked(1);
    }

    @Test
    public void testInitLockNeverAcquired() throws TimeoutException {
        when(artifactoryHome.isHaConfigured()).thenReturn(true);
        givenReleaseInitLockMarkerFileDoesntExist();

        haInitLock.runInsideInitLockIfNeeded(task, () -> false);

        verify(dbLocksService, times(0)).acquireLock(anyString(), anyString(), anyString(), anyLong(), any());
        verify(dbLocksService, times(0)).forceUnlock(any(), any());
        verifyTaskRunInvoked(1);
    }

    @SneakyThrows
    private void givenAcquireLockThrowsException() {
        doThrow(TimeoutException.class).when(dbLocksService)
                .acquireLock(anyString(), anyString(), anyString(), anyLong(), any(
                        TimeUnit.class));
    }

    private void givenDbLockedByOwner() {
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(dbLocksService.getLockInfoFromDbByCategoryAndLockKey(anyString(), anyString()))
                .thenReturn(Optional.of(new LockInfo(null, null, SERVER_ID, 0, null, 0)));
    }

    private void givenDbLockedByOther() {
        when(haNodeProperties.getServerId()).thenReturn(SERVER_ID);
        when(artifactoryHome.getHaNodeProperties()).thenReturn(haNodeProperties);
        when(dbLocksService.getLockInfoFromDbByCategoryAndLockKey(anyString(), anyString()))
                .thenReturn(Optional.of(new LockInfo(null, null, "Other", 0, null, 0)));
    }

    private void givenDbUnlocked() {
        when(dbLocksService.getLockInfoFromDbByCategoryAndLockKey(anyString(), anyString()))
                .thenReturn(Optional.empty());
    }

    private void givenReleaseInitLockMarkerFileExists() {
        when(artifactoryHome.getReleaseInitLockMarker()).thenReturn(releaseInitMarkerFile);
    }

    private void givenReleaseInitLockMarkerFileDoesntExist() {
        when(artifactoryHome.getReleaseInitLockMarker()).thenReturn(new File(""));
    }

    private void verifyReleaseInitLockMarkerFileDeleted() {
        Assert.assertFalse(releaseInitMarkerFile.exists());
    }

    @SneakyThrows
    private void verifyAcquireLockInvoked() {
        verify(dbLocksService, times(1))
                .acquireLock(anyString(), anyString(), eq(SERVER_ID), anyLong(), any(TimeUnit.class));
    }

    private void verifyForceUnlockInvoked() {
        verifyForceUnlockInvoked(2);
    }

    private void verifyForceUnlockInvoked(int timesInvoked) {
        verify(dbLocksService, times(timesInvoked)).forceUnlock(anyString(), anyString());
    }

    private void verifyTaskRunInvoked(int timesInvoked) {
        verify(task, times(timesInvoked)).run();
        verifyNoMoreInteractions(task);
    }
}