/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.aql.itest.service;

import com.google.common.collect.Sets;
import org.apache.commons.collections.CollectionUtils;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.AqlStatsCollectionService;
import org.artifactory.aql.api.AqlApiElement;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.model.AqlOperatorEnum;
import org.artifactory.aql.model.AqlOverloadSettings;
import org.artifactory.aql.model.AqlPermissionProvider;
import org.artifactory.aql.model.AqlRepoProvider;
import org.artifactory.aql.model.DomainSensitiveField;
import org.artifactory.aql.result.AqlComposedResult;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.AqlLazyResult;
import org.artifactory.aql.result.AqlRestResult;
import org.artifactory.aql.result.rows.AqlRowResult;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.storage.db.aql.dao.AqlDao;
import org.artifactory.storage.db.aql.itest.service.decorator.AqlQueryDecoratorContext;
import org.artifactory.storage.db.aql.itest.service.decorator.ArtifactRetentionColdSearchDecorator;
import org.artifactory.storage.db.aql.itest.service.decorator.DefaultSortDecorator;
import org.artifactory.storage.db.aql.itest.service.decorator.RemoteRepoDecorator;
import org.artifactory.storage.db.aql.itest.service.decorator.SupportBundlesRepoDecorator;
import org.artifactory.storage.db.aql.itest.service.decorator.TrashcanDecorator;
import org.artifactory.storage.db.aql.itest.service.decorator.VirtualRepoCriteriaDecorator;
import org.artifactory.storage.db.aql.parser.AqlParser;
import org.artifactory.storage.db.aql.parser.ParserElementResultContainer;
import org.artifactory.storage.db.aql.sql.builder.query.aql.AqlAdapter;
import org.artifactory.storage.db.aql.sql.builder.query.aql.AqlApiToAqlAdapter;
import org.artifactory.storage.db.aql.sql.builder.query.aql.AqlQuery;
import org.artifactory.storage.db.aql.sql.builder.query.aql.AqlQueryElement;
import org.artifactory.storage.db.aql.sql.builder.query.aql.OperatorQueryElement;
import org.artifactory.storage.db.aql.sql.builder.query.aql.ParserToAqlAdapter;
import org.artifactory.storage.db.aql.sql.builder.query.aql.SortDetails;
import org.artifactory.storage.db.aql.sql.builder.query.sql.SqlQuery;
import org.artifactory.storage.db.aql.sql.builder.query.sql.SqlQueryBuilder;
import org.artifactory.storage.db.aql.sql.result.AqlComposedResultImpl;
import org.artifactory.storage.db.aql.sql.result.AqlEagerResultImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.createWithEmptyResults;
import static org.artifactory.aql.api.internal.AqlBase.and;

/**
 * Execute the Aql queries by processing the three Aql steps one after the other:
 * Step 1 Convert the AqlApi or the parser result into AqlQuery.
 * Step 2 Convert the AqlQuery into SqlQuery.
 * Step 3 Execute the SqlQuery and return the results.
 *
 * @author Gidi Shabat
 */
@Service
public class AqlServiceImpl implements AqlService {
    private static final Logger log = LoggerFactory.getLogger(AqlServiceImpl.class);
    private final AqlPermissionProvider permissionProvider = new AqlPermissionProviderImpl();
    private final AqlRepoProvider repoProvider = new AqlRepoProviderImpl();
    @Autowired
    private AqlDao aqlDao;
    @Autowired
    private ArtifactoryDbProperties storageProperties;
    private AqlParser parser;
    private ParserToAqlAdapter parserToAqlAdapter;
    private AqlApiToAqlAdapter aqlApiToAqlAdapter;
    private SqlQueryBuilder sqlQueryBuilder;
    private AqlQueryOptimizer optimizer;
    private AqlQueryValidator validator;
    private AqlQueryDecorator externalDecorator; //Decorator used for user queries
    private AqlQueryDecorator internalDecorator; //Decorator used for internal api queries
    @Autowired
    private AqlQueryStreamService streamService;
    @Autowired
    private AqlStatsCollectionService aqlStatsCollectionService;

    @PostConstruct
    private void initDb() {
        // The parser is constructed by many internal elements therefore we create it once and then reuse it.
        // Please note that it doesn't really have state therefore we can use it simultaneously
        // TODO init the parser eagerly here not lazy
        parser = new AqlParser();
        parserToAqlAdapter = new ParserToAqlAdapter();
        sqlQueryBuilder = new SqlQueryBuilder();
        aqlApiToAqlAdapter = new AqlApiToAqlAdapter();
        optimizer = new AqlQueryOptimizer(storageProperties.getDbType());
        validator = new AqlQueryValidator();
        externalDecorator = new AqlQueryDecorator(new DefaultSortDecorator(), new TrashcanDecorator(),
                new SupportBundlesRepoDecorator(),
                new RemoteRepoDecorator(),
                new VirtualRepoCriteriaDecorator(),
                new ArtifactRetentionColdSearchDecorator());
        internalDecorator = new AqlQueryDecorator(new TrashcanDecorator(), new SupportBundlesRepoDecorator(),
                new VirtualRepoCriteriaDecorator());
    }

    /**
     * Converts the Json query into SQL query and executes the query eagerly
     */
    @Override
    public AqlEagerResult executeQueryEager(String query) {
        log.debug("Processing textual AqlApi query: {}", query);
        ParserElementResultContainer parserResult = parser.parse(query);
        return executeQueryEager(parserResult, null);
    }

    @Override
    public AqlEagerResult<AqlRowResult> executeQueryEager(String query,
            @Nonnull AqlOverloadSettings aqlOverloadSettings) {
        log.debug("Processing textual AqlApi query: {}", query);
        ParserElementResultContainer parserResult = parser.parse(query);
        return executeQueryEager(parserResult, aqlOverloadSettings);
    }

    /**
     * Converts the Json query into SQL query and executes the query lazy
     */
    @Override
    public AqlRestResult executeQueryLazy(String query) {
        log.debug("Processing textual AqlApi query: {}", query);
        return executeQueryLazy(query, null);
    }

    @Override
    public AqlRestResult executeQueryLazy(String query, AqlOverloadSettings aqlOverloadSettings) {
        ParserElementResultContainer parserResult = parser.parse(query);
        return executeQueryLazy(parserResult, aqlOverloadSettings);
    }

    /**
     * Converts the Json query into SQL query and executes the query lazy
     */
    @Override
    public <T extends AqlRowResult> AqlComposedResult executeQueryLazy(String originalMainQuery,
            AqlOperatorEnum operatorEnum,
            AqlBase<? extends AqlBase, T> extensionQuery) {
        log.debug("Processing textual AqlApi query {} and API AqlApi", originalMainQuery);
        ParserElementResultContainer parserResult = parser.parse(originalMainQuery);
        AqlQuery<AqlRowResult> originalMainAqlQuery = parserToAqlAdapter.toAqlModel(parserResult);
        AqlQuery<AqlRowResult> extensionAqlQuery = aqlApiToAqlAdapter.toAqlModel(extensionQuery);
        log.debug("Attempting to merge original query with the extension query");
        AqlQuery<AqlRowResult> mergedAqlQuery = merge(originalMainAqlQuery, extensionAqlQuery, operatorEnum);
        optimizer.optimize(mergedAqlQuery);
        validator.validate(mergedAqlQuery, permissionProvider);
        externalDecorator.decorate(mergedAqlQuery, new AqlQueryDecoratorContext(repoProvider, permissionProvider));
        log.trace("Successfully finished to convert the parser result into AqlApi query");
        return new AqlComposedResultImpl(this.streamService.getAqlQueryStreamResult(mergedAqlQuery),
                originalMainAqlQuery,
                extensionAqlQuery, mergedAqlQuery);
    }

    public <T extends AqlRowResult> AqlEagerResult<T> executeQueryEager(AqlBase<? extends AqlBase, T> aql,
            AqlOverloadSettings aqlOverloadSettings) {
        AqlQuery<T> aqlQuery = aqlApiToAqlAdapter.toAqlModel(aql);

        applySettingsOnAqlQuery(aqlOverloadSettings, aqlQuery);

        optimizer.optimize(aqlQuery);
        validator.validate(aqlQuery, permissionProvider);
        externalDecorator.decorate(aqlQuery, new AqlQueryDecoratorContext(repoProvider, permissionProvider));
        return getAqlQueryResult(aqlQuery);
    }

    /**
     * Converts the API's AqlApi query into SQL query and executes the query eagerly
     */
    @Override
    public <T extends AqlRowResult> AqlEagerResult<T> executeQueryEager(AqlBase<? extends AqlBase, T> aql) {
        AqlQuery<T> aqlQuery = aqlApiToAqlAdapter.toAqlModel(aql);
        optimizer.optimize(aqlQuery);
        internalDecorator.decorate(aqlQuery, new AqlQueryDecoratorContext(repoProvider, permissionProvider));
        return getAqlQueryResult(aqlQuery);
    }

    @Override
    public <T extends AqlRowResult> AqlLazyResult<T> executeQueryLazy(AqlBase<? extends AqlBase, T> aql) {
        log.debug("Processing API AqlApi query");
        AqlQuery<T> aqlQuery = aqlApiToAqlAdapter.toAqlModel(aql);
        optimizer.optimize(aqlQuery);
        internalDecorator.decorate(aqlQuery, new AqlQueryDecoratorContext(repoProvider, permissionProvider));
        return this.streamService.getAqlQueryStreamResult(aqlQuery);
    }

    /**
     * Converts the parser elements into AqlApi query, convert the AqlApi query to sql and executes the query eagerly
     */
    private <T extends AqlRowResult> AqlEagerResult<T> executeQueryEager(ParserElementResultContainer parserResult,
            AqlOverloadSettings aqlOverloadSettings) {
        log.trace("Converting the parser result into AqlApi query");
        AqlQuery<T> aqlQuery = parserToAqlAdapter.toAqlModel(parserResult);

        applySettingsOnAqlQuery(aqlOverloadSettings, aqlQuery);

        optimizer.optimize(aqlQuery);
        validator.validate(aqlQuery, permissionProvider);
        externalDecorator.decorate(aqlQuery, new AqlQueryDecoratorContext(repoProvider, permissionProvider,
                aqlOverloadSettings));
        log.trace("Successfully finished to convert the parser result into AqlApi query");
        return getAqlQueryResult(aqlQuery);
    }

    private <T extends AqlRowResult> void applySettingsOnAqlQuery(AqlOverloadSettings aqlOverloadSettings,
            AqlQuery<T> aqlQuery) {
        if (Objects.isNull(aqlOverloadSettings)) {
            return;
        }

        Long newLimit = aqlOverloadSettings.getLimit();
        if (Objects.nonNull(newLimit)) {
            aqlQuery.setLimit(newLimit);
        }

        SortDetails newSortDetails = getSortDetails(aqlOverloadSettings);
        if (Objects.nonNull(newSortDetails)) {
            aqlQuery.setSort(newSortDetails);
        }

        List<AqlApiElement<AqlApiItem>> queryElementsExtension = aqlOverloadSettings.getQueryElementsExtension();
        AqlOperatorEnum operatorEnum = aqlOverloadSettings.getQueryElementsExtensionOperatorEnum();
        if (!CollectionUtils.isEmpty(queryElementsExtension) && Objects.nonNull(operatorEnum)) {
            AqlApiItem extensionQuery =
                    createWithEmptyResults().filter(and(queryElementsExtension.toArray(AqlApiElement[]::new)));
            AqlQuery<AqlRowResult> extensionAqlQuery = aqlApiToAqlAdapter.toAqlModel(extensionQuery);
            aqlQuery.getAqlElements().add(new OperatorQueryElement(operatorEnum));
            aqlQuery.getAqlElements().add(AqlAdapter.open);
            aqlQuery.getAqlElements().addAll(extensionAqlQuery.getAqlElements());
            aqlQuery.getAqlElements().add(AqlAdapter.close);
        }

        List<DomainSensitiveField> newResultFields = aqlOverloadSettings.getResultFields();
        if (!CollectionUtils.isEmpty(newResultFields)) {
            aqlQuery.getResultFields().addAll(newResultFields.stream()
                            .filter(field -> !aqlQuery.getResultFields().contains(field))
                            .collect(Collectors.toList()));
        }
    }

    /**
     * Converts the parser elements into AqlApi query and executes the query lazy
     */
    private <T extends AqlRowResult> AqlRestResult executeQueryLazy(ParserElementResultContainer parserResult,
            AqlOverloadSettings aqlOverloadSettings) {
        log.trace("Converting the parser result into AqlApi query");
        AqlQuery<T> aqlQuery = parserToAqlAdapter.toAqlModel(parserResult);
        optimizer.optimize(aqlQuery);
        validator.validate(aqlQuery, permissionProvider);
        externalDecorator.decorate(aqlQuery, new AqlQueryDecoratorContext(repoProvider, permissionProvider,
                aqlOverloadSettings));
        log.trace("Successfully finished to convert the parser result into AqlApi query");
        return chooseStreamer(aqlQuery, aqlOverloadSettings);
    }

    private <T extends AqlRowResult> AqlRestResult chooseStreamer(AqlQuery<T> aqlQuery,
            AqlOverloadSettings aqlOverloadSettings) {
        if (Objects.nonNull(aqlOverloadSettings) && Objects.nonNull(aqlOverloadSettings.getStreamFunc())) {
            return this.streamService.getMultiStreamerWithStreamersFunctions(aqlQuery,
                    aqlOverloadSettings.getStreamFunc());
        }
        switch (aqlQuery.getQueryRepoType()) {
            case remote:
                AqlRestResult transitiveRemoteStreamer = this.streamService.getTransitiveRemoteStreamer(aqlQuery);
                aqlStatsCollectionService.collectRemoteSearchStats(transitiveRemoteStreamer);
                return transitiveRemoteStreamer;
            case virtual:
                AqlRestResult transitiveVirtualStreamer = this.streamService.getTransitiveVirtualStreamer(aqlQuery);
                aqlStatsCollectionService.collectRemoteSearchStats(transitiveVirtualStreamer);
                return transitiveVirtualStreamer;
            default:
                return this.streamService.getLocalStreamerWithWrapper(aqlQuery);
        }
    }

    /**
     * Converts the AqlApi query into SQL query and executes the query eagerly
     */
    private <T extends AqlRowResult> AqlEagerResult<T> getAqlQueryResult(AqlQuery<T> aqlQuery) {
        log.trace("Converting the AqlApi query into SQL query: {}", aqlQuery);
        SqlQuery<T> sqlQuery = sqlQueryBuilder.buildQuery(aqlQuery);
        log.trace("Successfully finished to convert the parser result into the following SQL query '{}'", sqlQuery);
        log.trace("processing the following SQL query: {}", sqlQuery);
        AqlEagerResultImpl<T> aqlQueryResult = aqlDao.executeQueryEager(sqlQuery, repoProvider);
        log.debug("Successfully finished to process SQL query with the following size: {}", aqlQueryResult.getSize());
        return aqlQueryResult;
    }

    private AqlQuery<AqlRowResult> merge(AqlQuery<AqlRowResult> originalMainQueryAqlQuery,
            AqlQuery<AqlRowResult> extensionAqlQuery, AqlOperatorEnum operatorEnum) {
        AqlQuery<AqlRowResult> merge = new AqlQuery<>();
        merge.setAction(originalMainQueryAqlQuery.getAction());
        merge.setDomain(originalMainQueryAqlQuery.getDomain());
        merge.setLimit(extensionAqlQuery.getLimit() != Long.MAX_VALUE ? extensionAqlQuery.getLimit() :
                originalMainQueryAqlQuery.getLimit());
        merge.setOffset(extensionAqlQuery.getOffset() != 0 ? extensionAqlQuery.getOffset() :
                originalMainQueryAqlQuery.getOffset());
        merge.setSort(extensionAqlQuery.getSort() != null ? extensionAqlQuery.getSort() :
                originalMainQueryAqlQuery.getSort());
        Set<DomainSensitiveField> mergedFields = Sets.newHashSet();
        mergedFields.addAll(originalMainQueryAqlQuery.getResultFields());
        mergedFields.addAll(extensionAqlQuery.getResultFields());
        merge.getResultFields().addAll(mergedFields);
        if (extensionAqlQuery.getAqlElements().isEmpty()) {
            merge.getAqlElements().addAll(originalMainQueryAqlQuery.getAqlElements());
        } else {
            merge.getAqlElements().add(AqlAdapter.open);
            merge.getAqlElements().addAll(originalMainQueryAqlQuery.getAqlElements());
            merge.getAqlElements().add(AqlAdapter.close);
            merge.getAqlElements().add(getOperator(operatorEnum));
            merge.getAqlElements().add(AqlAdapter.open);
            merge.getAqlElements().addAll(extensionAqlQuery.getAqlElements());
            merge.getAqlElements().add(AqlAdapter.close);
        }
        return merge;
    }

    private AqlQueryElement getOperator(AqlOperatorEnum operatorEnum) {
        switch (operatorEnum) {
            case or:
                return AqlAdapter.or;
            case and:
                return AqlAdapter.and;
            default:
                throw new UnsupportedOperationException("Aql merger support only And / or operations");
        }
    }

    @Nullable
    private SortDetails getSortDetails(AqlOverloadSettings aqlOverloadSettings) {
        if (Objects.isNull(aqlOverloadSettings.getSortField()) || Objects.isNull(aqlOverloadSettings.getSortType())) {
            return null;
        }
        SortDetails sortDetails = new SortDetails();
        sortDetails.setSortType(aqlOverloadSettings.getSortType());
        sortDetails.addField(aqlOverloadSettings.getSortField());
        return sortDetails;
    }
}