/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.artifacts.browse.treebrowser.tabs.builds;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.addon.build.artifacts.ProducedBy;
import org.artifactory.addon.build.artifacts.UsedBy;
import org.artifactory.api.jackson.JacksonFactory;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.ItemInfo;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.RepoPath;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import com.fasterxml.jackson.core.JsonGenerator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.ws.rs.core.StreamingOutput;
import java.io.IOException;
import java.util.stream.Stream;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class GetArtifactBuildsService implements RestService {
    private static final Logger log = LoggerFactory.getLogger(GetArtifactBuildsService.class);
    private RepositoryService repositoryService;
    private AddonsManager addonsManager;

    private ObjectMapper objectMapper = new ObjectMapper().setSerializationInclusion(JsonInclude.Include.NON_NULL);

    @Autowired
    public GetArtifactBuildsService(RepositoryService repositoryService,
            AddonsManager addonsManager) {
        this.repositoryService = repositoryService;
        this.addonsManager = addonsManager;
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        String path = request.getQueryParamByKey("path");
        String repoKey = request.getQueryParamByKey("repoKey");
        RepoPath repoPath = InternalRepoPathFactory.create(repoKey, path);
        if (repositoryService.isVirtualRepoExist(repoPath.getRepoKey())) {
            repoPath = repositoryService.getVirtualFileInfo(repoPath).getRepoPath();
        }
        FileInfo buildFileInfo = getFileInfo(repoPath);
        response.iModel(getArtifactBuildsAsStream(addonsManager.addonByType(BuildAddon.class), buildFileInfo.getSha1(),
                buildFileInfo.getMd5()));
    }

    public StreamingOutput getArtifactBuildsAsStream(BuildAddon buildAddon, String sha1, String md5) {
        return outputStream -> {
            try (outputStream) {
                JsonGenerator jsonGenerator = JacksonFactory.createGenerator(outputStream, false);
                jsonGenerator.writeStartObject();
                jsonGenerator.writeArrayFieldStart("usedBy");

                try (Stream<UsedBy> usedByStream = buildAddon.retrieveDependantBuilds(sha1, md5)) {
                    usedByStream.forEach(usedBuild -> writeToStream(jsonGenerator, usedBuild));
                } catch (Exception e) {
                    log.error("Failed to iterate dependent builds {}", e.getMessage());
                }

                jsonGenerator.writeEndArray();
                jsonGenerator.writeArrayFieldStart("producedBy");

                try (Stream<ProducedBy> producedByStream = buildAddon.retrieveProducedBuilds(sha1, md5)) {
                    producedByStream.forEach(producedBuild -> writeToStream(jsonGenerator, producedBuild));
                } catch (Exception e) {
                    log.error("Failed to iterate containing builds {}", e.getMessage());
                }

                jsonGenerator.writeEndArray();
                jsonGenerator.writeEndObject();
                jsonGenerator.close();
            }
        };
    }

    protected void writeToStream(JsonGenerator jsonGenerator, ProducedBy build) {
        try {
            jsonGenerator.writeObject(build);
            jsonGenerator.flush();
        } catch (IOException e) {
            log.error("Failed writing object to stream {}", e.getMessage());
        }
    }

    protected void writeToStream(JsonGenerator jsonGenerator, UsedBy build) {
        try {
            objectMapper.writeValue(jsonGenerator, build);
            jsonGenerator.flush();
        } catch (IOException e) {
            log.error("Failed writing object to stream {}", e.getMessage());
        }
    }

    /**
     * get File info by repo path
     *
     * @param repoPath - repo path
     * @return file info instance
     */
    private FileInfo getFileInfo(RepoPath repoPath) {
        ItemInfo fileInfo = repositoryService.getItemInfo(repoPath);
        return (FileInfo) fileInfo;
    }
}
